import expetator.bundle as bundle
import expetator.monitoring_csv as monitoring_csv
import expetator.monitoring_list as monitoring_list
import expetator.watermark as watermark
import sys

def remove_watermark(target_file, target_dir):

    bundle_data, zip_fid = bundle.init_bundle(target_file)

    try:
        moj = monitoring_csv.read_bundle_csv('mojitos', bundle_data, zip_fid)
        moj_cleaned = watermark.remove_watermark_blocks(moj, frequency=20)
    except:
        moj_cleaned = None

    try:
        power = monitoring_list.read_bundle_list('power', bundle_data, zip_fid)
        power_cleaned = watermark.remove_watermark_blocks(power, frequency=20)
    except:
        power_cleaned = None

    watermark.remove_wt_name(bundle_data)
    if target_file.endswith('.zip'):
        target_file = target_file[:-4]

    bundle.save_bundle(target_file, bundle_data, target_dir)

    if not moj_cleaned is None:
        monitoring_csv.write_bundle_csv('mojitos', bundle_data, moj_cleaned, target_dir)
    if not power_cleaned is None:
        monitoring_list.write_bundle_list('power', bundle_data, power_cleaned, target_dir)

def main():
    if len(sys.argv) != 3:
        print('Usage : %s main_file target_directory')
        sys.exit(0)

    target_file = sys.argv[1]
    target_dir = sys.argv[2]
    remove_watermark(target_file, target_dir)

if __name__ == '__main__':
    main()



#! /usr/bin/python3

import sys
import scipy.integrate as integrate
from statistics import mean, median

import expetator.bundle as bundle
import expetator.monitoring_list as monitoring_list

import matplotlib.pyplot as plt

def add_energy(basename, bundle_data=None, zip_fid=None):
    if bundle_data is None:
        bundle_data, zip_fid = bundle.init_bundle(basename)

    power = monitoring_list.read_bundle_list('power', bundle_data, zip_fid)
    
    energy = []
    delta = []
    mean_power = []
    med_power = []
    
    for experiment in power:
        expe_energy = 0
        expe_delta = 0
        expe_power = []
        try:
            for power_profile in experiment:
                tmp_delta = max(power_profile['#timestamp']) - min(power_profile['#timestamp'])
                expe_delta = max(expe_delta, tmp_delta)
        
                expe_energy += integrate.trapz(power_profile['power'], x=power_profile['#timestamp'])
                
                expe_power.extend(power_profile['power'])
        except:
            pass
        if expe_power == []:
            expe_energy = -1
            expe_delta = -1
            expe_power = [-1]

        energy.append(expe_energy)
        delta.append(expe_delta)
        mean_power.append(mean(expe_power))
        med_power.append(median(expe_power))
        
    bundle_data['energy'] = energy
    bundle_data['time'] = delta
    bundle_data['mean_power'] = mean_power
    bundle_data['median_power'] = med_power

    return(bundle_data)

def main():
    basename = sys.argv[1]
    bundle_data = add_energy(basename)

    bundle_data.to_csv(basename+".csv", sep=' ', index=False)

if __name__ == "__main__":
    main()



from expetator.monitors.kwollect import Power as Kwollect
from expetator.monitors.lperf import Lperf
from expetator.monitors.mojitos import Mojitos
from expetator.monitors.powergpu import Power as PowerGpu



#TODO: put watermark detection at delta=0 if corr<.6

# For watermark detection
from sklearn import decomposition
import numpy as np


# for demo (to remove ?)
import matplotlib.pyplot as plt


# Tools for watermark detection
def df_to_vector(df):
    np.seterr(invalid='ignore')
    try:
        tmp = df.drop('#timestamp', axis=1)
    except:
        tmp = df
    pca = decomposition.PCA(n_components=1)
    pca.fit(tmp)
    x = pca.transform(tmp)
    return x.flatten()

def get_wm_ref(total_time=30, nb_steps=2, freq=10):
    low = int(total_time/(2*nb_steps+1))
    high = (total_time-low*(nb_steps+1))/nb_steps

    mark = np.array([])
    for _ in range(nb_steps):
        mark=np.append(mark, [-1]*int(low*freq))
        mark=np.append(mark, [1]*int(high*freq))
    mark=np.append(mark, [-1]*int(low*freq))
    return mark


def get_shift(dataframe, frequency=10, duration=30, backward=False):
    # Takes the best shift using a 10x frequency
    precision = 4
    timeframe = np.arange(0, duration, 1/(frequency*precision))

    # Gets the watermark on the timeframe
    watermark = get_wm_ref(freq=frequency*precision)

    # Gets the data regularly interpolated in the right window
    ts = dataframe['#timestamp'].to_numpy()
    if backward:
        pos_in_data = np.searchsorted(ts, max(ts)-duration )
        tmp_ts = ts[pos_in_data:]
        tmp_data = df_to_vector(dataframe[pos_in_data:])     
    else:
        pos_in_data = np.searchsorted(ts, min(ts)+duration )
        tmp_ts = ts[:pos_in_data+1]
        tmp_data = df_to_vector(dataframe[:pos_in_data+1])     
    
    data = np.interp(timeframe+min(tmp_ts), tmp_ts, tmp_data)
    
    deltamax = len(timeframe)//(2*precision)
    current = None
    res = None
    for delta in range(-deltamax, deltamax+1):
        if delta <= 0:
            coeff = np.corrcoef(data[:len(timeframe)+delta] ,
                                watermark[-delta:])[0][1]
        else:
            coeff = np.corrcoef(data[delta:] ,
                                watermark[:len(timeframe)-delta])[0][1]
        
        
        if current is None or abs(coeff) > current:
            current=abs(coeff)
            if backward:
                res = np.searchsorted(ts, max(ts)-duration+delta/(precision*frequency) ), coeff
            else:
                res = np.searchsorted(ts, min(ts)+duration+delta/(precision*frequency) ), coeff
        
            #print(delta, res)

    if np.isnan(current) or current < .6:
        res = pos_in_data, 0
    
    return res

def remove_watermark(dataframe, frequency=10, duration=30):
    if len(dataframe) == 0:
        return dataframe
    delta_deb, conf_deb = get_shift(dataframe, frequency, duration)
    delta_fin, conf_fin = get_shift(dataframe, frequency, duration, backward=True)
    return dataframe[delta_deb:delta_fin+1].reset_index(drop=True)

def remove_watermark_blocks(block, frequency=10, duration=30):
    return [
        [
            remove_watermark(block[experiment][host], frequency, duration)
            for host in range(len(block[experiment]))
        ]
        for experiment in range(len(block))
    ]


def remove_wt_name(dataframe, target='fullname', signature='wt-30-'):
    s_len = len(signature)
    dataframe[target] = [elem[s_len:] for elem in dataframe[target]]

## Tool for virtualisation

def demo_watermark_detection(focus, freq=10):
    df = focus.loc[:, focus.columns != '#timestamp']
    norm_focus=(df-df.min())/(df.max()-df.min())
    norm_focus['#timestamp'] = focus['#timestamp']
    fig, ax = plt.subplots()
    norm_focus.plot(x='#timestamp', figsize=(10,6), ax=ax)

    delta_deb, conf_deb = get_shift(focus, freq)
    delta_fin, conf_fin = get_shift(focus, freq, backward=True)

    # the actual data
    norm_focus[delta_deb:delta_fin+1].plot(x='#timestamp', linewidth=4, ax=ax)

    # common part of the watermark
    watermark_reference = (get_wm_ref(freq=freq)+1)/2
    nb_watermark = len(watermark_reference)

    #start
    w_start = min(focus[delta_deb:delta_fin+1]['#timestamp'])
    plt.plot(np.linspace(w_start-30, w_start-1/freq,len(watermark_reference)),
             watermark_reference)
    plt.vlines(w_start, 0, 1)
    plt.vlines(w_start-1/freq, 0, 1)

    #end
    w_end = max(focus[delta_deb:delta_fin+1]['#timestamp'])
    plt.plot(np.linspace(w_end + 1/freq, w_end+30,len(watermark_reference)),
             watermark_reference)
    plt.vlines(w_end, 0, 1)
    plt.vlines(w_end+1/freq, 0, 1)

    print(w_end-w_start)    



from expetator.benchmarks.cpubench import CpuBench
from expetator.benchmarks.genericbench import GenericBench
from expetator.benchmarks.gpucpubench import GpuCpuBench
from expetator.benchmarks.gpumembench import GpuMemBench
from expetator.benchmarks.gromacsbench import GromacsBench
from expetator.benchmarks.membench import MemBench
from expetator.benchmarks.mpibench import MpiBench
from expetator.benchmarks.npbbench import NpbBench
from expetator.benchmarks.percentagebench import PercentageBench
from expetator.benchmarks.sleepbench import SleepBench
from expetator.benchmarks.watermark import WaterMark



import os

source_script = """#!/usr/bin/python3

import signal
import os

def signal_handler(signal_number, frame):
    os.close(fd)

fd = os.open("/dev/cpu_dma_latency", os.O_RDWR)
os.write(fd, b'\\0\\0\\0\\0')

signal.signal(signal.SIGINT, signal_handler)
signal.pause()
"""

terminate_script = """#!/bin/bash
ps aux | grep script_nodeepsleep.py | grep sudo | awk '{print $2}' | xargs sudo-g5k kill
"""

class Nodeepsleep:
    'Activates or deactivates CX'
    def __init__(self, modes = {True, False}):
        self.modes = modes
        self.binary = '/tmp/bin/script_nodeepsleep.py'
        self.terminate = '/tmp/bin/terminate_nodeepsleep.py'
        self.cur_mode = False
        
    def build(self, executor):
        'Builds the software'
        self.executor = executor
        with open(self.binary, 'w') as out_file:
            out_file.write(source_script)
        os.chmod(self.binary, 0o777)

        with open(self.terminate, 'w') as out_file:
            out_file.write(terminate_script)
        os.chmod(self.terminate, 0o777)
        
    def available_states(self):
        'Returns all the modes'
        return self.modes

    def start(self, mode):
        'Sets the right mode on all hosts'
        self.curr_mode = mode
        if mode:
            self.executor.hosts(self.binary+' &', root=True)

    def stop(self, output_file=None):
        'Stops the script'
        if self.curr_mode:
            self.executor.hosts(self.terminate, root=True)
        self.curr_mode = False
            
    def get_state(self):
        'Returns the current state'
        return self.curr_mode

    def state_to_str(self):
        'Returns the current state as a string'
        return '%s' % self.curr_mode

    def get_labels(self):
        'Returns the label for the C0 state'
        return ('C0only', )



#! /usr/bin/python3

import os
import sys
import pandas as pd

def get_nb_freq(file_list):
    m_size = 0
    representative = ''

    for filename in file_list:
        stat = os.stat(filename)
        if stat.st_size > m_size:
            m_size = stat.st_size
            representative = filename

    data = pd.read_csv(representative, sep=' ')
    return len(set(data['fmax']))

def main():
    print(get_nb_freq(sys.argv[1:]))

if __name__ == '__main__':
    main()




import os
import pandas as pd
import matplotlib.pyplot as plt
import sys


def _read_csv(filename):
    df = pd.read_csv(filename, sep=' ', skipinitialspace=True)
    if df.columns[-1].startswith('Unnamed'):
        df.drop(columns=df.columns[-1:], axis=1, inplace=True)
    return df



def read_host_csv(prefix, hostname, startTime, basename, fullname, archive_fid=None):
    fullpath= '%s_%s/%s_%s_%s' % (basename, prefix, hostname, fullname, startTime)
    if archive_fid is None:
        if not os.path.exists(fullpath):
            fullpath = '%s_%s/%s_%s_%s' % (basename, prefix, 'localhost', fullname, startTime)
        with open(fullpath) as file_id:
            data = _read_csv(fullpath)
    else:
        if not fullpath in archive_fid.namelist():
            fullpath = '%s_%s/%s_%s_%s' % (basename, prefix, 'localhost', fullname, startTime)
        with archive_fid.open(fullpath) as file_id:
            data = _read_csv(file_id)
    data = data.dropna(axis='columns')
    return data

def read_run_csv(prefix, hostname, startTime, basename, fullname, hostlist, archive_fid=None):
    return [read_host_csv(prefix, host, startTime, basename, fullname, archive_fid) for host in hostlist.split(';')]
       
def read_bundle_csv(prefix, bundle, archive_fid=None):
    'Reads mojitO/S-like files associated to a bundle'
    csv_data = [read_run_csv(prefix, row.hostname, row.startTime, row.basename, row.fullname, row.hostlist, archive_fid) 
                    for index, row in bundle.iterrows()]
    return csv_data




def write_host_csv(prefix, hostname, startTime, basename, fullname, data, target_directory):
    fullpath= '%s/%s_%s/%s_%s_%s' % (target_directory, basename, prefix, hostname, fullname, startTime)
    os.makedirs('%s/%s_%s' % (target_directory, basename, prefix), exist_ok=True)

    data.to_csv(fullpath,sep=' ', index=False)


def write_run_csv(prefix, hostname, startTime, basename, fullname, hostlist, data, target_directory):
    for index, host in enumerate(hostlist.split(';')):
        write_host_csv(prefix, host, startTime, basename, fullname, data[index], target_directory)


def write_bundle_csv(prefix, bundle, data, target_directory):
    for index, row in bundle.iterrows():
        write_run_csv(prefix, row.hostname, row.startTime, row.basename, row.fullname, row.hostlist, data[index], target_directory) 

def show_csv(filename, norm=False):
    filename = sys.argv[-1]

    a = _read_csv(filename)
    a['#timestamp'] = a['#timestamp']-a['#timestamp'][0]
    if norm:
        tmp = (a/a.max())
        tmp['#timestamp'] = a['#timestamp']
        a = tmp
    a.plot(x='#timestamp')
    plt.show(block=True)

def show_csv_main():
    filename = sys.argv[-1]
    norm = len(sys.argv) == 3
    show_csv(filename, norm)



#! /usr/bin/python3

import sys
import pandas as pd

def clean_csv(filename, nb):

    df = pd.read_csv(filename, sep=' ', )

    # remove the last incomplete experiment
    last = (len(df)//nb)*nb
    df = df[:last]

    # remove complete experiment is the monitoring failed
    for bloc_id in range(len(df)//nb-1,-1, -1):
        for idx in range(nb):
            tmp = df.iloc[bloc_id*nb+idx]
            if min(tmp[10:]) <= -1:
                df = df.drop( range(bloc_id*nb, (bloc_id+1)*nb) )
                break
                

    df.to_csv(filename[:-4]+"_cleaned.csv", sep=' ', index=False)

def main():
    if len(sys.argv) != 3:
        print("usage %s filename expe_size\nwhere expe_size is the number of experiment for one experimental set")
    
    else:
        clean_csv(sys.argv[1], int(sys.argv[2]))

if __name__ == '__main__':
    main()

    



import pandas as pd
import numpy as np
import json
import os
import zipfile

def init_bundle(bundlename):
    'Reads an experiment file'
    if bundlename.endswith('.zip'):
        zip_fid = zipfile.ZipFile(bundlename, 'r')
        bundlename = bundlename[:-4]
        experiments = pd.read_csv(zip_fid.open(bundlename), sep=' ')
    else:    
        zip_fid = None
        experiments = pd.read_csv(bundlename, sep=' ')
    
    experiments['basename'] = bundlename
    return experiments, zip_fid

def save_bundle(bundle_name, bundle_data, target_directory):
    os.makedirs(target_directory, exist_ok=True)
    bundle_data.drop('basename', axis=1).to_csv(target_directory+'/'+bundle_name,sep=' ', index=False)



def merge_timeseries_blocks(references, additions, prefix = 'add_', key='#timestamp'):
    return [
        [
            merge_timeseries(references[experiment_id][host_id],
                             additions[experiment_id][host_id],prefix, key)
            for host_id in range(len(references[experiment_id]))
        ]
        for experiment_id in range(len(references))
    ]

def merge_timeseries(reference, addition, prefix = 'add_', key='#timestamp'):
    reference = reference.copy()
    
    delta = reference[key][0] - addition[key][0]
    intermediate_timestamps = reference[key]-delta
    
    for column in addition:
        if column != key:
            new_name = prefix+addition[column].name
            
    
            data = np.interp(intermediate_timestamps, addition[key], addition[column])

            reference[new_name] =data

    return reference

    delta_ref = reference[key][0]
    span_ref = max(reference[key]) - reference[key][0]
    delta_add = addition[key][0]
    span_add = max(addition[key]) - addition[key][0]
    
    print(delta_ref, span_ref)
    print(delta_add, span_add)

def normy(focus):
    df = focus.loc[:, focus.columns != '#timestamp']
    norm_focus=(df-df.min())/(df.max()-df.min())
    norm_focus['#timestamp'] = focus['#timestamp']
    return norm_focus
    



#! /usr/bin/python3

import expetator.monitoring_csv as monitoring_csv
import expetator.monitoring_list as monitoring_list
import expetator.watermark as watermark
import sys
import pandas as pd
import matplotlib.pyplot as plt
import json

if __name__ == '__main__':
    if len(sys.argv) != 3 or not sys.argv[1] in ['csv', 'list']:
        print("""Usage :
%s csv file.csv

or

%s list file.list
""")
        sys.exit(0)

    filename = sys.argv[2]
    if sys.argv[1] == 'csv':
        
        a = monitoring_csv._read_csv(filename)
        watermark.demo_watermark_detection(a, 10)
        plt.show(block=True)


    if sys.argv[1] == 'list':

        with open(filename) as f_id:
            data = json.load(f_id)

            for name, timestamps, power in data:

                df = pd.DataFrame([timestamps, power]).transpose()
                df.columns = ["#timestamp", "Values"]

                watermark.demo_watermark_detection(df, 10)
                plt.show(block=True)



        
                



from expetator.leverages.dvfs import Dvfs
from expetator.leverages.gpuclock import GpuClock
from expetator.leverages.gpupow import GpuPower
from expetator.leverages.neosched import NeoSched
from expetator.leverages.nodeepsleep import Nodeepsleep
from expetator.leverages.pct import Pct
from expetator.leverages.powercap import Powercap
from expetator.leverages.template import Template



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="expetator",
    version="0.3.23",
    author="Georges Da Costa",
    author_email="georges.da-costa@irit.fr",
    description="A framework for monitoring HPC applications using DVFS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.irit.fr/sepia-pub/expetator",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={'': ['benchmarks/genericbench/*', 'benchmarks/*tpr', 'benchmarks/*.c', 'benchmarks/*.tgz', 'benchmarks/*.cu',
                       'monitors/*.bz2', 'monitors/*.diff',
                       'leverages/*.[ch]', 'leverages/*.sh', 'leverages/*_mak']},
    include_package_data=True,
    install_requires=['execo', 'requests'],
    entry_points={
        'console_scripts': [
            'remove_watermark = expetator.remove_watermark:main',
            'csv_plot = expetator.monitoring_csv:show_csv_main',
            'list_plot = expetator.monitoring_list:show_list_main',
            'add_energy = expetator.add_energy:main',
            'get_nb_freq = expetator.get_nb_freq:main',
            'clean_csv = expetator.clean_csv:main',
        ]
    }


)



import os
import sys
import json
import pandas as pd
import matplotlib.pyplot as plt

## Power
def read_run_list(prefix, hostname, startTime, basename, fullname, hostlist=None, archive_fid=None):
    fullpath= '%s_%s/%s_%s_%s' % (basename, prefix, hostname, fullname, startTime)
    result = []

    if archive_fid is None:
        with open(fullpath) as file_id:
            raw_data = json.loads(file_id.read())
    else:
        with archive_fid.open(fullpath) as file_id:
            raw_data = json.loads(file_id.read())
    
    data = {host:(timestamp, values) for (host, timestamp, values) in raw_data}

    for host in hostlist.split(';'):
        name, _ = host.split('.', maxsplit=1)
        df = pd.DataFrame(list(data[name])).transpose()
        df.columns = ["#timestamp", prefix]
        result.append(df)

    return result

def read_bundle_list(prefix, bundle, archive_fid=None):
    'Reads the power files associated to a bundle'    
    list_data = [read_run_list(prefix, row.hostname, row.startTime, row.basename, row.fullname, row.hostlist, archive_fid) 
                    for index, row in bundle.iterrows()]
    return list_data





def write_run_list(prefix, hostname, startTime, basename, fullname, hostlist, data, target_directory):
    fullpath= '%s/%s_%s/%s_%s_%s' % (target_directory, basename, prefix, hostname, fullname, startTime)
    os.makedirs('%s/%s_%s' % (target_directory, basename, prefix), exist_ok=True)

    hosts = hostlist.split(';')
    res = []

    for index, host in enumerate(hosts):
        host = host.split('.')[0]
        tmp = [host, list(data[index]['#timestamp']), list(data[index][prefix])]
        res.append(tmp)

    with open(fullpath, 'w') as file_id:
        json.dump(res, file_id)
    


def write_bundle_list(prefix, bundle, data, target_directory):
    'Writes the power files associated to a bundle'    
    for index, row in bundle.iterrows():
        write_run_list(prefix, row.hostname, row.startTime, row.basename, row.fullname, row.hostlist, data[index], target_directory)


def show_list(filename):
    with open(filename) as f_id:
        data = json.load(f_id)

        for name, timestamps, power in data:
            plt.plot([t-timestamps[0] for t in timestamps],
                     power, label=name)
        
    plt.legend()
    plt.show(block=True)



def show_list_main():
    filename = sys.argv[-1]
    show_list(filename)
